# Copyright 2014 Max Sharples
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to manage set up and running of Hivejobs
"""
import sys
import logging

from optparse import Option
#from optparse import OptionError
#from optparse import OptionGroup
from optparse import OptionParser

from apiarist.emr import EMRRunner

log = logging.getLogger(__name__)

# sentinel value; used when running HiveJob as a script
_READ_ARGS_FROM_SYS_ARGV = '_READ_ARGS_FROM_SYS_ARGV'

class HiveJobLauncher(object):
    
    OPTION_CLASS = Option
    
    def __init__(self, job_name, args=None):

        self.job_name = job_name

        # TODO _ allow argument to be passed in 
        # to be used to compose the script (variables/parameters)
        self.passthrough_options = []

        self.option_parser = OptionParser(usage=self._usage(),
                                        option_class=self.OPTION_CLASS,
                                        add_help_option=False)
        self.configure_options()

        # arguments provided on command line
        if args==_READ_ARGS_FROM_SYS_ARGV:
            self._cl_args = sys.argv[1:]
            self.options, args = self.option_parser.parse_args(self._cl_args)
        else:
            self.options, args = self.option_parser.parse_args(args)

        # after named args are removed, only remaining arg is the source of data
        self.input_data = args[0]
            
    def execute(self):
        """Run the job
        """
        print("Launching job {0}".format(self.job_name))
        # log the options being used
        print self.emr_job_runner_kwargs()
        # self.set_up_logging(quiet=self.options.quiet, verbose=self.options.verbose, stream=self.stderr)
        with self.make_runner() as runner:
            runner.run()

    def make_runner(self):
        """Make a runner based on arguments provided"""
        # TODO  when we need to make other types of runer (local)
        return EMRRunner(**self.emr_job_runner_kwargs())

    def emr_job_runner_kwargs(self):
        slave_instance_type = self.options.slave_instance_type
        master_instance_type = self.options.master_instance_type or slave_instance_type
        return {
                'input_path': self.input_data,
                'output_dir': self.options.output_dir,
                'hive_query': self.hive_query(),
                'scratch_dir': self.options.scratch_dir,
                'job_name': self.job_name,
                'master_instance_type': master_instance_type,
                'slave_instance_type': slave_instance_type,
                'num_instances': self.options.num_instances,
                'ami_version': self.options.ami_version,
                'hive_version': self.options.hive_version,
                }

    def hive_query(self):
        # implemented in subclass HiveJob
        raise NotImplementedError

    def configure_options(self):
        """Define the arguments for this script
        """
        # TODO move EMR-specific options to the emr module
        # TODO allow YAML config file
        self.option_parser.add_option(
                '--output-dir', dest='output_dir', action='store', default=False
                )
        self.option_parser.add_option(
                '--s3-scratch-uri', dest='scratch_dir', action='store', default=False
                )
        self.option_parser.add_option(
                '--ec2-master-instance-type', dest='master_instance_type',
                action='store', default=False
                )
        self.option_parser.add_option(
                '--ec2-instance-type', dest='slave_instance_type',
                action='store', default='m3.xlarge'
                )
        self.option_parser.add_option(
                '--num-ec2-instances', dest='num_instances',
                action='store', default=2
                )
        self.option_parser.add_option(
                '--ami-version', dest='ami_version', 
                action='store', default='latest'
                )
        self.option_parser.add_option(
                '--hive-version', dest='hive_version', 
                action='store', default='latest'
                )

    @classmethod
    def _usage(cls):
        """Command line usage string for this class"""
        return ("usage: python [script path|executable path|--help]")




# Copyright 2014 Max Sharples
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from apiarist.s3 import *

class HiveQuery(object):
    """
    Class to manage the contruction of the Hive query including the boilerplate to load data
    Can return a Hive query for local or EMR execution.
    """
    HIVE_TYPES = [
            'TINYINT',  # 1 byte
            'SMALLINT'  # 2 byte
            'INT'       # 4 byte
            'BIGINT'    # 8 byte
            'FLOAT'     # 4 byte (single precision floating point numbers)
            'DOUBLE'    # 8 byte (double precision floating point numbers)
            'BOOLEAN'   # you know what these are
            'STRING',   # up to 2GB 
            ]
            # NOTE: There are no date data types; dates are treated as strings.
            # There are several date functions which operate on strings.

    def __init__(self, table_name, input_columns, output_columns, query):
        self.table_name         = table_name
        self.results_table_name = table_name + "_results"
        self.query              = self._parse_query(query)
        self.input_columns      = input_columns
        self.output_columns     = output_columns
        if query[-1:] != ';':
            raise ValueError, "query must terminate with a semi-colon"
        if table_name not in query:
            raise ValueError, "query does not contain a reference to the table"
        # TODO validate the input/output columns for proper data types and reserved keywords

    def __repr__(self):
        return "HiveQuery:{}...".format(self.query[:80])

    def _csv_serde_jar(self):
        """Using a JAR for serialisation/deserialisation in the Hive tables
        """
        # FIXME this method doesn't feel at home here
        try:
            # known location on S3
            serde = os.environ["CSV_SERDE_JAR_S3"]
        except KeyError:
            # ensure the jar is up on S3
            jar_loc = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'jars', 'csv-serde-1.1.2-0.11.0-all.jar'))
            jar_path = os.environ['S3_BASE_PATH'] + 'jars/csv-serde.jar'
            upload_file_to_s3(jar_loc, jar_path)
            os.environ["CSV_SERDE_JAR_S3"] = serde = jar_path
        return serde

    def _parse_query(self, query):
        """Condense spaces"""
        return re.sub(r"\s+"," ",query)

    def emr_hive_script(self, data_source, output_dir, temp_table_dir):
        """Generate the complete Hive script for EMR
        outputs a comma-delimited file via a hive textfile table
        """
        # boilerplate
        parts = [
            "ADD JAR {0};".format(self._csv_serde_jar()), # 
            "SET hive.exec.compress.output=false;"
            ]
        # add the table in which we'll load the source data
        parts += self._create_table_ddl(self.table_name, self.input_columns, temp_table_dir)
        # add statement to load the source data into this table
        parts.append("LOAD DATA INPATH '{0}' INTO TABLE {1};".format(data_source, self.table_name)) 
        # add a table to select the results into (for CSV formatting)
        parts += self._create_table_ddl(self.results_table_name, self.output_columns, output_dir)
        # insert the results of the supplied query into this table
        parts.append("INSERT INTO TABLE {0} {1}".format(self.results_table_name, self.query))
        # and finally, the query
        parts.append(self.query)
        # return a string that can be written to a file and run on Hive
        return "\n".join(parts)

    def _create_table_ddl(self, name, columns, location):
        """Create a Hive table to store CSV data
        """
        return [
            "CREATE EXTERNAL TABLE {0} ({1})".format(name, self._column_ddl(columns)), 
            "ROW FORMAT serde 'com.bizo.hive.serde.csv.CSVSerde'",
            "STORED AS TEXTFILE", 
            "LOCATION '{0}';".format(location)
            ]

    def _column_ddl(self, columns):
        """Get column defintions for Hive table
        """
        cols = ["`{0}` {1}".format(col[0], col[1]) for col in columns]
        return ", ".join(cols)
 



# Copyright 2014 Max Sharples
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to inherit your HiveJobs from. See README for more info
"""
import logging

from apiarist.launch import HiveJobLauncher
from apiarist.script import HiveQuery
from apiarist.launch import _READ_ARGS_FROM_SYS_ARGV

log = logging.getLogger(__name__)

class HiveJob(HiveJobLauncher):

    def __init__(self, args=None):
        super(HiveJob, self).__init__(self._job_name(), args)

    def hive_query(self):
        """Get the Hive script object based on provided params
        """
        return HiveQuery(self.table(), self.input_columns(), self.output_columns(), self.query())

    def _job_name(self):
        return self.__class__.__name__

    @classmethod
    def run(cls):
        """Entry point for running job from the command-line.
        """
        hive_job = cls(args=_READ_ARGS_FROM_SYS_ARGV)
        hive_job.execute()

    ## methods which define the query logic and I/O ##

    def input_columns(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError

    def output_columns(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError

    def table(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError

    def query(self):
        """Create this in your HiveJob subclass"""
        raise NotImplementedError



# Copyright 2014 Max Sharples
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" S3/storage utility methods """

import os
import re
from boto.s3.connection import S3Connection
from boto.s3.key import Key


def copy_s3_file(source, destination):
    """ Copy an S3 object from one location to another
    """
    dest_bucket, dest_key = parse_s3_uri(destination)
    source_bucket, source_key = parse_s3_uri(source)
    conn = S3Connection(os.environ['AWS_ACCESS_KEY_ID'], os.environ['AWS_SECRET_ACCESS_KEY'])
    bkt = conn.get_bucket(dest_bucket)
    return bkt.copy_key(dest_key, source_bucket, source_key)

def upload_file_to_s3(file_path, s3_path):
    """Create an S3 object from the contents of a local file
    """
    s3_bucket, s3_key = parse_s3_uri(s3_path)
    conn = S3Connection(os.environ['AWS_ACCESS_KEY_ID'], os.environ['AWS_SECRET_ACCESS_KEY'])
    bkt = conn.get_bucket(s3_bucket)
    k = Key(bkt)
    k.key = s3_key
    return k.set_contents_from_filename(file_path)       

def parse_s3_uri(uri):
    """Parse an S3 uri from: s3://bucketname/some/other/path/info/
    to: 
        bucket = bucketname
        key = some/other/path/info
    """
    m = re.search(r'(s3://)([A-Za-z0-9_-]+)/(\S*)', uri)
    if m:
        return (m.group(2),m.group(3))
    else:
        return None






# Copyright 20014 Max Sharples
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
try:
    from setuptools import setup
    # arguments that distutils doesn't understand
    setuptools_kwargs = {
        'install_requires': [
            'boto>=2.2.0'
        ],
        'provides': ['apiarist']
    }
except ImportError:
    from distutils.core import setup
    setuptools_kwargs = {}

import apiarist

setup(
    author='Max Sharples',
    author_email='maxsharples@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Distributed Computing',
    ],
    description='Python Hive query framework',
    license='Apache',
    long_description=open(os.path.abspath(os.path.join(os.path.dirname(__file__),'README.md'))).read(),
    name='apiarist',
    packages=[
        'apiarist'
    ],
    url='http://github.com/msharp/apiarist',
    version=apiarist.__version__,
    **setuptools_kwargs
)
